/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DatabaseModel.java,v 1.2 2001/08/20 09:23:38 max Exp $
 *
 */

package make.datamodel.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.DatabaseMetaData;
import java.util.Hashtable;
import java.util.Enumeration;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import make.util.Debug;
import make.util.Strings;

/**
 * DatabaseModel is a wrapper for a Connection object
 * that supports connection pooling, and some helper functions.
 */
public class DatabaseModel {
    /*CONST****************************/
    public static final byte DB_UNKNOWN = 1;
    public static final byte DB_MYSQL = 1;
    public static final byte DB_ORACLE = 2;
    public static final byte DB_POSTGRESQL = 3;
    public static final int DATETIME = Strings.DATETIME;
    public static final int DATE = Strings.DATE;
    public static final int TIME = Strings.TIME;
    private static final DateFormat[] mysqlDateFormat = {
	new SimpleDateFormat("yyyy-MM-dd HH:mm"),
	new SimpleDateFormat("yyyy-MM-dd"),
	new SimpleDateFormat("HH:mm"),
    };
    private static final DateFormat[] oracleDateFormat = {
	new SimpleDateFormat("'TO_DATE'(''yyyy-MM-dd HH:mm'',' ''YYYY-MM-DD HH24:MI''')"),
	new SimpleDateFormat("'TO_DATE'(''yyyy-MM-dd'',' ''YYYY-MM-DD''')"),
	new SimpleDateFormat("'TO_DATE'(''HH:mm'',' ''HH24:MI''')"),
    };
    /*properties***************************************/
    private ConnectionFactory connectionFactory;
    private Connection[] connections;
    private Hashtable threads = new Hashtable();
    private int nextConnection = 0;
    private boolean limitClause = false;
    private byte db = DB_UNKNOWN;
    private DateFormat[] dateFormat = mysqlDateFormat;
    /*Constructor*************************************/
    public DatabaseModel(int poolSize, ConnectionFactory _cf) {
	connectionFactory = _cf;
	connections = new Connection[poolSize];
	// find out db engine
	/*if (url.indexOf("mysql")!=-1)
	  db = DB_MYSQL;
	  else if (url.indexOf("oracle")!=-1)
	  db = DB_ORACLE;
	  else if (url.indexOf("postgr")!=-1)
	  db = DB_POSTGRESQL;*/
	//
	try {
	    connections[nextConnection] = connectionFactory.createConnection();
	    DatabaseMetaData metaData = connections[nextConnection].getMetaData();
	    System.err.println("NOTICE: opened database "+
			       metaData.getDatabaseProductName());
	    // TODO: find out which database engine this is
	    db = DB_MYSQL;
	    // some workarounds for different SQL syntax
	    limitClause = db==DB_MYSQL || db==DB_POSTGRESQL;
	    if (db==DB_ORACLE)
		dateFormat = oracleDateFormat;
	} catch (SQLException e) {
	    e.printStackTrace();
	}
    }
    public DatabaseModel(int poolSize, String _driver, String _url,
			 String _usr, String _pw) {
	this(poolSize, new DefaultConnectionFactory(_driver, _url, _usr, _pw));
    }
    public DatabaseModel(int poolSize, String _driver, String _url) {
	this(poolSize, _driver, _url, null, null);
    }
    public void close() {
	try {
	    for (int z=0; z<connections.length; z++)
		if (connections[z]!=null)
		    connections[z].close();
	} catch (SQLException e) {
	    e.printStackTrace();
	} finally {
	    connections = null;
	}
    }
    /*Info***********************************/
    public byte getDatabaseType() {
	return db;
    }
    public boolean hasLimitClause() {
	return limitClause;
    }
    public DateFormat getDateFormat(int nr) {
	return dateFormat[nr];
    }
    /*Pool**********************************/
    private int nextIndex() {
	final int idx = nextConnection;
	nextConnection++;
	if (nextConnection>=connections.length)
	    nextConnection = 0;
	return idx;
    }
    private Connection getConnection2() throws SQLException {
	return getConnection(nextIndex());
    }
    private Connection getConnection(int idx) throws SQLException {
	if (connections[idx]==null)
	    reconnect(idx);
	return connections[idx];
    }
    private int getNumber(Thread thread) {
	Integer idx = (Integer)threads.get(thread);
	/*if (Debug.DEBUG && idx==null)
	  System.err.println("ERROR: no idx for thread "+thread);*/
	return idx==null ? -1 : idx.intValue();
    }
    protected void garbageCollection() {
	for (Enumeration enum=threads.keys();
	     enum.hasMoreElements();) {
	    Thread t = (Thread)enum.nextElement();
	    if (!t.isAlive())
		threads.remove(t);
	}
    }
    public final Connection getConnection() throws SQLException {
	garbageCollection();
	Thread t = Thread.currentThread();
	int idx = getNumber(t);
	if (idx==-1) {
	    idx = nextIndex();
	    threads.put(t, new Integer(idx));
	    //System.err.println(t+" => "+idx);
	}
	return getConnection(idx);
    }
    public final Statement createStatement() throws SQLException {
	return getConnection().createStatement();
    }
    public final PreparedStatement prepareStatement(String sql) throws SQLException {
	final long t0 = System.currentTimeMillis();
	PreparedStatement ps = getConnection().prepareStatement(sql);
	if (db==DB_MYSQL)
	    ps.setEscapeProcessing(false);
	final long t1 = System.currentTimeMillis();
	final long td = t1-t0;
	if (td>500) {
	    System.err.println("WARNING: slow database "+td+" ms");
	    System.err.println("SQL="+sql);
	}
	return ps;
    }
    /*Query********************************/
    public boolean tableExists(String name) throws SQLException {
	Statement s = createStatement();
	try {
	    s.executeQuery("SELECT COUNT(*) FROM "+name);
	} catch (SQLException e) {
	    return false;
	} finally {
	    s.close();
	}
	return true;
    }
    /*Connection**************************/
    public synchronized void reconnect(int z) throws SQLException {
	synchronized(this) {
	    //System.err.println("Reconnecting "+z);
	    if (connections[z]!=null) {
		try {
		    connections[z].close();
		} catch (SQLException e) {
		}
		connections[z] = null;
		System.gc();
	    }
	    connections[z] = connectionFactory.createConnection();
	    //System.err.println("connections["+z+"] = "+connections[z]);
	}
    }
    public synchronized void reconnect(Thread t) throws SQLException {
	reconnect(getNumber(t));
    }
    public synchronized void reconnect() throws SQLException {
	reconnect(Thread.currentThread());
    }
}
