/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: RootDatabaseModel.java,v 1.4 2001/12/14 17:09:08 max Exp $
 *
 */

package make.datamodel.db;

import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;

/**
 * RootDatabaseModel is an object which dynamically manages multiple 
 * database objects which are loaded from a central database. This is 
 * a very special class, you probably will not need it when writing 
 * simple database applications.
 */
public abstract class RootDatabaseModel extends DatabaseModel
    implements Runnable {
    /*properties***************************************/
    protected Thread thread = new Thread(this, "RootDatabase");
    protected Hashtable dbs = new Hashtable();
    protected Timestamp lastFind;
    protected String fields = "name,driver,usr,pwd,url";
    protected PropertyChangeSupport pcs
        = new PropertyChangeSupport(this);
    /*Constructor*************************************/
    public RootDatabaseModel(ConnectionFactory _cf, String _fields) {
        super(1, _cf);
        if (_fields!=null)
            fields = _fields;
        thread.setDaemon(true);
    }
    public RootDatabaseModel(String _driver, String _url, 
                             String _usr, String _pwd, String _fields) {
        super(1, _driver, _url, _usr, _pwd);
        if (_fields!=null)
            fields = _fields;
        thread.setDaemon(true);
    }
    public RootDatabaseModel(String _driver, String _url, String _fields) {
        this(_driver, _url, null, null, _fields);
    }
    public void start() {
        try {
            reload();
            lastFind = new Timestamp(System.currentTimeMillis());
        } catch (SQLException e) {
            e.printStackTrace();
        }
        thread.setDaemon(true);
        thread.start();
    }
    public void close() {
        thread.interrupt(); thread = null;
        for (Enumeration enum=dbs.keys(); enum.hasMoreElements();) {
            String name = (String)enum.nextElement();
            DatabaseModel db = getDatabase(name);
            dbs.remove(name);
            pcs.firePropertyChange(name, db, null);
            db.close();
        }
        dbs = null;
        pcs = null;
    }
    /*Info***********************************/
    public DatabaseModel getDatabase(String name) {
        return (DatabaseModel)dbs.get(name);
    }
    public Enumeration getDatabaseNames() {
        return dbs.keys();
    }
    /*Beans************************************/
    public void addDatabaseChangeListener(PropertyChangeListener pcl) {
        pcs.addPropertyChangeListener(pcl);
    }
    public void removeDatabaseChangeListener(PropertyChangeListener pcl) {
        pcs.removePropertyChangeListener(pcl);
    }
    /*Abstract**********************************/
    protected abstract DatabaseModel loadDatabase(ResultSet rs) 
        throws SQLException;
    /*Internal*********************************/
    protected synchronized void reload() throws SQLException {
        synchronized(dbs) {
            Vector names = new Vector(dbs.size());
            for (Enumeration enum=dbs.keys(); enum.hasMoreElements();)
                names.addElement(enum.nextElement());
            Statement s = createStatement();
            ResultSet rs = s.executeQuery("SELECT "+fields+" FROM dbs");
            while (rs.next()) {
                String name = rs.getString("name");
                DatabaseModel newValue = loadDatabase(rs);
                DatabaseModel oldValue = newValue==null
                    ?(DatabaseModel)dbs.get(name)
                    :(DatabaseModel)dbs.put(name, newValue);
                if (oldValue!=null)
                    names.removeElement(name);
                pcs.firePropertyChange(name, oldValue, newValue);
            }
            rs.close();
            s.close();
            for (Enumeration enum=names.elements(); enum.hasMoreElements();) {
                String name = (String)enum.nextElement();
                DatabaseModel oldValue = (DatabaseModel)dbs.remove(name);
                pcs.firePropertyChange(name, oldValue, null);
            }
        }
    }
    /*Runnable**********************************/
    public void run() {
        for (int counter=0;;counter++) {
            try {
		// check for new databases
                PreparedStatement findNewDatabases = prepareStatement
                    ("SELECT "+fields+" FROM dbs WHERE last_modified>?");
                findNewDatabases.setTimestamp(1, lastFind);
                ResultSet rs = findNewDatabases.executeQuery();
                while (rs.next()) {
                    String name = rs.getString("name");
                    DatabaseModel newValue = loadDatabase(rs);
                    DatabaseModel oldValue = newValue==null
                        ?(DatabaseModel)dbs.get(name)
                        :(DatabaseModel)dbs.put(name, newValue);
                    pcs.firePropertyChange(name, oldValue, newValue);
                }
                rs.close();
                findNewDatabases.close();
                lastFind = new Timestamp(System.currentTimeMillis());
		// check for deleted databases
                if ((counter%4)==0) {
                    Statement s = createStatement();
                    rs = s.executeQuery("SELECT COUNT(*) FROM dbs");
                    rs.next();
                    if (rs.getInt(1)!=dbs.size())
                        reload();
                    rs.close();
                    s.close();
                }
            } catch (SQLException e) {
                e.printStackTrace();
                try {
                    reconnect();
                } catch (SQLException e2) {
                    e2.printStackTrace();
                }
            }
            try {
                thread.sleep(30000);
            } catch (InterruptedException e) {
                break;
            }
        }
    }
}
