/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: MemoryEmbeddedManager.java,v 1.7 2001/10/11 21:32:19 max Exp $
 *
 */

package make.datamodel.embedded;

import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.Hashtable;
import java.util.Enumeration;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import make.io.Streams;
import make.xml.DOMUtil;

/**
 * EmbeddedManager implementation that saves all its data
 * in memory in a hash containing byte arrays. 
 * You may use this class to create new documents which are not
 * yet saved, i.e. it is unknown where they will be saved later.
 *
 * This class implements java.io.Serializable, that means
 * you can save it to disk using native Java serialization.
 */
public class MemoryEmbeddedManager 
    implements EmbeddedManager, Serializable {
    /*CONST***********************************/
    public static final String DEFAULT_NAME = "[RAM]";
    /*properties*******************************/
    protected Hashtable entries = new Hashtable();
    protected Node document;
    protected String name;
    /*Constructor*****************************/
    /**
     * Creates a new instance.
     * @param _doc the initial document
     * @param _name the name of this EmbeddedManager
     */
    public MemoryEmbeddedManager(Node _doc, String _name) { 
	document = _doc;
	name = _name;
    }
    /**
     * Creates a new instance.
     * @param _doc the initial document
     */
    public MemoryEmbeddedManager(Node _doc) { 
	this(_doc, DEFAULT_NAME);
    }
    /**
     * Creates a new instance.
     * @param _name the name of this EmbeddedManager
     */
    public MemoryEmbeddedManager(String _name) { 
	this(null, _name);
    }
    /**
     * Creates a new instance.
     */
    public MemoryEmbeddedManager() { 
	this(null, DEFAULT_NAME);
    }
    /*Object*********************************/
    public String toString() { return name; }
    /*EmbeddedManager***************************/
    public String[] getNames() throws IOException {
	String[] names = new String[entries.size()];
	int z=0;
	for (Enumeration enum=entries.keys(); enum.hasMoreElements();)
	    names[z++] = (String)enum.nextElement();
	return names;
    }
    public Hashtable getSizes() throws IOException {
	Hashtable sz = new Hashtable();
	for (Enumeration enum=entries.keys(); enum.hasMoreElements();) {
	    String name = (String)enum.nextElement();
	    sz.put(name, new Long(((byte[])entries.get(name)).length));
	}
	return sz;
    }
    public boolean contains(String name) { 
	return entries.containsKey(name); 
    }
    public void replaceEntry(String name, InputStream data)
	throws IOException {
	if (data==null)
	    entries.remove(name);
	else {
	    ByteArrayOutputStream os = new ByteArrayOutputStream();
	    Streams.copy(data, os);
	    entries.put(name, os.toByteArray());
	}
    }
    public InputStream getEntry(String name) throws IOException {
	byte[] data = (byte[])entries.get(name);
	return data==null?null:new ByteArrayInputStream(data);
    }
    public void replaceDocument(Node n) throws IOException {
	document = n;
    }
    public Node getDocument() throws IOException, SAXException {
	return document;
    }
    /*Serialization****************************/
    private void writeObject(java.io.ObjectOutputStream out)
	throws IOException {
	out.writeByte(1);
	out.writeUTF(name);
	out.writeObject(entries);
	byte[] data = document==null
	    ?null
	    :DOMUtil.toByteArray(document.getOwnerDocument(), "utf-8");
	out.writeObject(data);
    }
    private void readObject(java.io.ObjectInputStream in)
	throws IOException, ClassNotFoundException {
	in.readByte();
	name = in.readUTF();
	entries = (Hashtable)in.readObject();
	byte[] data = (byte[])in.readObject();
	try {
	    document = data==null
		?null
		:DOMUtil.parse(data, "utf-8").getDocumentElement();
	} catch (SAXException e) {
	    throw new IOException(e.getLocalizedMessage());
	}
    }
}
