/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ModuleClassLoader.java,v 1.6 2001/08/20 09:23:39 max Exp $
 *
 */

package make.module;

import java.util.Hashtable;
import java.io.File;
import java.io.InputStream;
import java.net.URL;
import make.util.Debug;

/**
 * ModuleClassLoader is an adapter that converts a
 * ResourceLoader object into a ClassLoader which can be
 * used to dynamically load Java classes at runtime.
 * Note that this class is the direct opposite of 
 * ClassLoaderResourceLoader.
 */
public class ModuleClassLoader extends ClassLoader {
    /*properties***************************/
    protected ClassLoader parent;
    protected ResourceLoader resourceLoader;
    protected Hashtable cache = new Hashtable();
    /*Constructor****************************/
    /**
     * Creates a new ClassLoader instance.
     * @param _parent the ClassLoader to fallback to
     * @param _rl the ResourceLoader object
     */
    public ModuleClassLoader(ClassLoader _parent, ResourceLoader _rl) {
	parent = _parent;
	resourceLoader = _rl;
    }
    /*ClassLoader***************************/
    protected Class loadClass(String name, boolean resolve) 
	throws ClassNotFoundException {
	// first search in cache
	Class cls2 = (Class)cache.get(name);
	if (cls2!=null) {
	    if (resolve)
		resolveClass(cls2);
	    return cls2;
	}
	// system/parent class?
	try {
	    Class cls = parent==null
		?findSystemClass(name)
		:parent.loadClass(name);
	    if (resolve)
		resolveClass(cls);
	    return cls;
	} catch (Exception e) {
	}
	// now try to load from resourceLoader
        String classFileName =
            name.replace('.', '/') + ".class";
	byte[] classData
	    = resourceLoader.loadResource(classFileName);
	if (classData!=null) {
	    try {
		Class cls = defineClass(name, classData,
					0, classData.length);
		if (resolve)
		    resolveClass(cls);
		cache.put(name, cls);
		return cls;
	    } catch(Throwable t) {
		Debug.report(t);
	    }
	}
	throw new ClassNotFoundException(name);
    }
    public InputStream getResourceAsStream(String name) {
	InputStream is = parent==null
	    ?getSystemResourceAsStream(name)
	    :parent.getResourceAsStream(name);
	if (is!=null)
	    return is;
	return resourceLoader.getResourceAsStream(name);
    }
    public URL getResource(String name) {
	URL url = parent==null
	    ?getSystemResource(name)
	    :parent.getResource(name);
	if (url!=null)
	    return url;
	return resourceLoader.getResource(name);
    }
}
