/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ZipResourceLoader.java,v 1.10 2001/08/20 09:23:39 max Exp $
 *
 */

package make.module;

import java.io.File;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import make.util.Debug;
import make.io.Streams;

/**
 * ZipResourceLoader loads resources from a ZIP or a JAR file.
 */
public class ZipResourceLoader 
    implements ResourceLoader {
    /*properties*******************************/
    protected ZipFile zipFile;
    protected File file;
    /*Constructor******************************/
    /**
     * Creates a new ResourceLoader instance
     * @param _f the ZIP file to open
     */
    public ZipResourceLoader(File _f) throws IOException {
	file = _f;
	zipFile = new ZipFile(file);
    }
    public String toString() {
	return "[ZipResourceLoader file="+file+"]";
    }
    /**
     * Finds an entry in the ZIP file.
     */
    protected ZipEntry getEntry(String name) {
	ZipEntry ze;
	ze = zipFile.getEntry(name);
	if (ze!=null) return ze;
	while (!Character.isJavaIdentifierPart(name.charAt(0)))
	    name = name.substring(1);
	ze = zipFile.getEntry(name);
	if (ze!=null) return ze;
	ze = zipFile.getEntry("/"+name);
	if (ze!=null) return ze;
	ze = zipFile.getEntry("./"+name);
	if (ze!=null) return ze;
	return null;
    }
    /**
     * Destructs this object and closes all files.
     * It is optional to call this method, and in very rare cases
     * it is useful.
     */
    public void close() {
	try {
	    zipFile.close();
	} catch (IOException e) {
	}
	zipFile = null;
	file = null;
    }
    /*ResourceLoader***************************/
    public URL getResource(String name) {
	ZipEntry ze = getEntry(name);
	if (ze!=null)
	    try {
		while (!Character.isJavaIdentifierPart(name.charAt(0)))
		    name = name.substring(1);
		return new URL("jar:file:" + file.getAbsolutePath() + "!/" + name);
	    } catch(MalformedURLException e) {
		Debug.report(e);
	    }
	return null;
    }
    public InputStream getResourceAsStream(String name) {
	ZipEntry ze = getEntry(name);
	if (ze!=null)
	    try {
		return zipFile.getInputStream(ze);
	    } catch(IOException e) {
		Debug.report(e);
	    }
	return null;
    }
    public byte[] loadResource(String name) {
	ZipEntry ze = getEntry(name);
	if (ze!=null)
	    try {
		long size = ze.getSize();
		if (size==-1) {
		    InputStream is = zipFile.getInputStream(ze);
		    ByteArrayOutputStream os = new ByteArrayOutputStream();
		    Streams.copy(is, os);
		    is.close();
		    byte[] data = os.toByteArray();
		    os.close();
		    return data;
		} else {
		    if (size>(64*1024*1024))
			return null;
		    InputStream is = zipFile.getInputStream(ze);
		    byte[] data = new byte[(int)size];
		    Streams.read(is, data);
		    is.close();
		    return data;
		}
	    } catch(IOException e) {
		Debug.report(e);
	    }
	return null;
    }
}
