/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: TrusteeTask.java,v 1.5 2001/08/20 09:23:39 max Exp $
 *
 */

package make.scheduler;

import java.util.Date;
import make.util.Debug;

/**
 * TrusteeTask is an adapter that converts a Trustee
 * object into a Task object - which can be used in Scheduler.
 */
public class TrusteeTask implements Task {
    /*properties*************************/
    private Trustee trustee;
    private long next;
    /*Constructor**********/
    /**
     * Creates a new Task instance.
     * @param _t the trustee to be converted
     */
    public TrusteeTask(Trustee _t) {
	trustee = _t;
	next = System.currentTimeMillis();
    }
    public String toString() {
	return "[TrusteeTask trustee="+trustee+"]";
    }
    /*Info**************/
    /**
     * Returns the associated Trustee object
     */
    public Trustee getTrustee() {
	return trustee;
    }
    /*TrusteeTask**************/
    /**
     * Handles an exception caught after trustee.run().
     * The default implementation dumps the exception to System.err.
     * You may want to override this method.
     */
    protected void handleException(Exception e) {
	Debug.report(e, "in trustee "+trustee);
    }
    /*Task*************/
    public Date getNextTime() {
	return new Date(next);
    }
    public synchronized void run() {
	boolean busy = false;
	try {
	    busy = trustee.run();
	} catch (Exception t) {
	    handleException(t);
	} catch (Error e) {
	    Debug.report(e, "in trustee "+trustee);
	    throw e;
	}
	final long delay = busy
	    ?trustee.getBusyDelay()
	    :trustee.getStandbyDelay();
	final long now = System.currentTimeMillis();
	next = now + delay;
    }
}
