/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DefaultInputLinkFactory.java,v 1.28 2002/01/03 12:23:06 max Exp $
 *
 */

package make.swing.il;

import java.util.Map;
import java.util.List;
import java.util.Collection;
import java.util.Locale;
import javax.swing.BoxLayout;
import org.w3c.dom.Element;
import make.util.Debug;
import make.xml.DOMUtil;
import make.datamodel.slot.Slot;
import make.datamodel.slot.SlotContainer;
import make.datamodel.slot.LocalizedSlotContainer;
import make.datamodel.slot.LocalizedSlotContainerProxy;
import make.swing.field.Field;
import make.swing.field.TextField;
import make.swing.field.TextAreaField;
import make.swing.field.TranslateTextField;
import make.swing.field.PasswordField;
import make.swing.field.IntegerField;
import make.swing.field.CheckBoxField;
import make.swing.field.RadioField;
import make.swing.field.DateField;
import make.swing.field.SelectField;
import make.swing.field.ListField;
import make.swing.field.MenuSelectField;
import make.swing.field.TimeOffsetComboField;

public class DefaultInputLinkFactory implements InputLinkFactory {
    /*properties*****************************/
    protected SlotContainer slotContainer;
    protected LocalizedSlotContainer localizedSlotContainer;
    protected InputLink localizedInputLink;
    protected LocaleSelect localeSelect;

    /*Constructor******************************/
    public DefaultInputLinkFactory(SlotContainer _sc) {
        slotContainer = _sc;
    }

    /*Info********************************/
    public SlotContainer getSlotContainer() {
        return slotContainer;
    }

    /*Misc*************************************/
    protected Object getList(Element vd, Locale locale) {
        return null;
    }
    protected Slot getSlot(Element vd) {
        String path = DOMUtil.getElementParameterNS(vd, XML_NAMESPACE, "path", false);
        if (path==null)
            path = DOMUtil.getElementParameterNS(vd, XML_NAMESPACE, "name", false); 
        if (path==null)
            return null;
        return slotContainer.getSlot(path);
    }
    protected Field createField(Element vd, Locale locale) {
        String type = getType(vd);
        if (type.equals("text")) {
            try {
                int columns = 20;
                String columnsString = vd.getAttributeNS(XML_NAMESPACE, "columns");
                if (columnsString!=null && !"".equals(columnsString))
                    columns = Integer.parseInt(columnsString);
                return new TextField(columns);
            } catch (NumberFormatException e) {
                Debug.report(e);
                return null;
            }
        }
        if (type.equals("translatetext")) {
            try {
                int columns = 20;
                Map map = (Map)getList(vd, locale);
                String columnsString = vd.getAttributeNS(XML_NAMESPACE, "columns");
                if (columnsString!=null && !"".equals(columnsString))
                    columns = Integer.parseInt(columnsString);
                return new TranslateTextField(columns, map);
            } catch (NumberFormatException e) {
                Debug.report(e);
                return null;
            }
        }
        if (type.equals("textarea")) {
            try {
                int rows = Integer.parseInt(vd.getAttributeNS(XML_NAMESPACE, "rows"));
                int columns = Integer.parseInt(vd.getAttributeNS(XML_NAMESPACE, "columns"));
                return new TextAreaField(rows, columns);
            } catch (NumberFormatException e) {
                Debug.report(e);
                return null;
            }
        }
        if (type.equals("menuselect")) {
            Map map = (Map)getList(vd, locale);
            Object[] menu = (Object[])getList
                (DOMUtil.getUniqueTagNS(vd, XML_NAMESPACE, "menu"), locale);
            Element fvd = DOMUtil.getUniqueTagNS(vd, XML_NAMESPACE, "field");
            Field f = createField(DOMUtil.getChildElementsNS(fvd, XML_NAMESPACE)[0], locale);
            return new MenuSelectField(f, map, menu);
        }
        if (type.equals("password"))
            return new PasswordField();
        if (type.equals("integer"))
            return new IntegerField(true);
        if (type.equals("select")) {
            Object list = getList(vd, locale);
            if (list instanceof Map) {
                Map map = (Map)list;
                return new SelectField(map);
            } else {
                Collection collection = (Collection)list;
                return new SelectField(collection);
            }
        } 
        if (type.equals("multiselect")) {
            Object list = getList(vd, locale);
            if (list instanceof Map) {
                Map map = (Map)list;
                return new ListField(map, ";");
            } else {
                Collection collection = (Collection)list;
                return new ListField(collection, ";");
            }
        } 
        if (type.equals("checkbox")) {
            String caption = DOMUtil.getLocalizedElementParameterNS
                (vd, XML_NAMESPACE, "caption", locale, false);
            if (caption==null)
                caption = DOMUtil.getLocalizedElementStringNS
                    (vd, XML_NAMESPACE, "caption", locale, false);
            return new CheckBoxField(caption);
        }
        if (type.equals("radio")) {
            String axisString = DOMUtil.getElementParameterNS
                (vd, XML_NAMESPACE, "axis", false);
            int axis = axisString==null || !axisString.equals("y")
                ? BoxLayout.X_AXIS : BoxLayout.Y_AXIS;
            Element[] buttons = DOMUtil.getChildElementsNS
                (vd, XML_NAMESPACE, "radiobutton");
            Object[] values = new Object[buttons.length];
            String[] captions = new String[buttons.length];
            for (int z=0; z<buttons.length; z++) {
                values[z] = DOMUtil.getElementParameterNS
                    (buttons[z], XML_NAMESPACE, "value", false);
                captions[z] = DOMUtil.getLocalizedElementParameterNS
                    (buttons[z], XML_NAMESPACE, "caption", locale, false);
                if (captions[z]==null)
                    captions[z] = ""+values[z];
            }
            return new RadioField(axis, values, captions);
        }
        if (type.equals("datetime"))
            return new DateField(true, 0);
        if (type.equals("date"))
            return new DateField(true, 1);
        if (type.equals("time"))
            return new DateField(true, 2);
        if (type.equals("timeoffset")) {
            List list = (List)getList(vd, locale);
            return new TimeOffsetComboField((String[])list.toArray(new String[]{}));
        }
        return null;
    }
    public String getType(Element vd) {
        return vd.getLocalName();
    }
    /*InputLinkFactory**************************/
    public InputLink createInputLink(Element vd, Locale locale) {
        String type = getType(vd);
        Element childrenElement
            = DOMUtil.getUniqueTagNS(vd, XML_NAMESPACE, "children");
        Element[] childrenElements
            = childrenElement==null
            ?null
            :DOMUtil.getChildElementsNS(childrenElement, XML_NAMESPACE);
        if (type.equals("label"))
            return new Label(DOMUtil.getElementString(vd));
        if (type.equals("button"))
            return new Button(vd, locale);
        if (type.equals("glue"))
            return new Glue();
        if (type.equals("panel"))
            return new PanelInputLink(vd, childrenElements, locale, this);
        if (type.equals("tabbed"))
            return new TabbedInputLink(vd, childrenElements, locale, this);
        if (type.equals("grid"))
            return new GridInputLink(vd, childrenElements, locale, this);
        if (type.equals("vbox"))
            return new BoxInputLink(vd, BoxLayout.Y_AXIS,
                                    childrenElements, locale, this);
        if (type.equals("hbox"))
            return new BoxInputLink(vd, BoxLayout.X_AXIS,
                                    childrenElements, locale, this);
        if (type.equals("dialog"))
            return new DialogInputLink(vd, locale, this);
        if (type.equals("scroll"))
            return new ScrollPaneInputLink(vd, locale, this);
        if (type.equals("localize")) {
            // lock this object...
            synchronized(this) {
                // save old values
                final InputLink oldLocIL = localizedInputLink;
                final LocalizedSlotContainer oldLocSlotC = localizedSlotContainer;
                final LocaleSelect oldLocSel = localeSelect;
                final SlotContainer oldSlotC = slotContainer;
                localizedSlotContainer = new LocalizedSlotContainerProxy(slotContainer, null);
                localeSelect = null;
                slotContainer = localizedSlotContainer;
                try {
                    Element child = DOMUtil.getUniqueTagNS(vd, XML_NAMESPACE, "child");
                    Element[] chs = DOMUtil.getChildElementsNS(child, XML_NAMESPACE);
                    localizedInputLink = new ProxyInputLink(null);
                    InputLink ret = createInputLink(chs[0], locale);
                    ((ProxyInputLink)localizedInputLink).setInputLink(ret);
                    if (localeSelect!=null)
                        localeSelect.reload();
                    return ret;
                } finally {
                    // restore old vlaues
                    localizedInputLink = oldLocIL;
                    localizedSlotContainer = oldLocSlotC;
                    localeSelect = oldLocSel;
                    slotContainer = oldSlotC;
                }
            }
        }
        if (type.equals("locale")) {
            if (localizedInputLink==null) {
                System.err.println("WARNING: creating locale InputLink without localizedInputLink");
                return null;
            }
            localeSelect = new LocaleSelect(vd, localizedSlotContainer,
                                            localizedInputLink, locale);
            return localeSelect;
        }
        Field fld = createField(vd, locale);
        if (fld!=null) {
            if ("false".equals(vd.getAttributeNS(XML_NAMESPACE, "enabled")))
                fld.setEnabled(false);
            return new FieldInputLink(vd, getSlot(vd), fld, locale);
        }
        return null;
    }
}
