/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DefaultEditorKit.java,v 1.9 2001/10/30 17:06:07 max Exp $
 *
 */

package make.swing.text;

import java.io.IOException;
import java.text.BreakIterator;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.UnsupportedFlavorException;
import javax.swing.Action;
import javax.swing.AbstractAction;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Position;
import javax.swing.text.Segment;
import javax.swing.undo.UndoManager;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.CannotRedoException;
import javax.swing.event.UndoableEditListener;
import javax.swing.event.UndoableEditEvent;
import make.util.Debug;
import make.swing.ButtonManager;
import make.swing.panel.InterfacePanel;
import make.swing.panel.DialogInterfacePanel;
import make.swing.text.view.ViewFactory;

public abstract class DefaultEditorKit 
    implements EditorKit, PropertyChangeListener, 
               ChangeListener, UndoableEditListener,
               ClipboardOwner {
    /*CONST***********************************/
    public static final boolean DUMP_CLIPBOARD = false;

    /*properties*****************************************/
    protected TextComponent editor;
    protected UndoManager undo = new UndoManager();
    protected DragAndDropSupport dragAndDropSupport = new DragAndDropSupport(this);
    protected String currentFormat = null;
    private final Action defaultAction = new DefaultKeyTypedAction();

    /*Constructor*************************/
    public DefaultEditorKit() {}

    /*DefaultEditorKit*******************************************/
    protected void installActions(ButtonManager bm) {
        bm.installActions(actions);
    }
    protected void uninstallActions(ButtonManager bm) {
        bm.uninstallActions(actions);
    }
    public void beep() {
        if (editor!=null)
            editor.getToolkit().beep();
    }
    public String getCurrentFormat() {
        return currentFormat;
    }
    public void setCurrentFormat(String _cf) {
        currentFormat = _cf;
    }
    public TextComponent getTextComponent() {
        return editor;
    }
    public Clipboard getClipboard() {
        try {
            return editor.getToolkit().getSystemClipboard();
        } catch (SecurityException e) {
            return null;
        }
    }

    /*PropertyChangeListener****************************/
    public void propertyChange(PropertyChangeEvent evt) {
        if (evt.getSource()==editor)
            if (evt.getPropertyName().equals(editor.BUTTON_MANAGER)) {
                ButtonManager ov = (ButtonManager)evt.getOldValue();
                if (ov!=null)
                    uninstallActions(ov);
                ButtonManager nv = (ButtonManager)evt.getNewValue();
                if (nv!=null)
                    installActions(nv);
            } else if (evt.getPropertyName().equals(editor.DOCUMENT)) {
                Document oldValue = (Document)evt.getOldValue();
                if (oldValue!=null)
                    oldValue.removeUndoableEditListener(this);
                Document newValue = (Document)evt.getNewValue();
                if (newValue!=null)
                    newValue.addUndoableEditListener(this);
            } else if (evt.getPropertyName().equals(editor.CARET)) {
                Caret oldValue = (Caret)evt.getOldValue();
                if (oldValue!=null)
                    oldValue.addChangeListener(this);
                Caret newValue = (Caret)evt.getNewValue();
                if (newValue!=null)
                    newValue.addChangeListener(this);
            }
    }

    /*UndoableEditListener***************************/
    public void undoableEditHappened(UndoableEditEvent e) {
        undo.addEdit(e.getEdit());
        undoActionInstance.update();
        redoActionInstance.update();
    }

    /*EditorKit*******************************************/
    public void install(TextComponent c) {
        if (Debug.DEBUG && editor!=null)
            throw new InternalError();
        editor = c;
        ButtonManager bm = editor.getButtonManager();
        if (bm!=null)
            installActions(bm);
        Document doc = editor.getDocument();
        if (doc!=null)
            doc.addUndoableEditListener(this);
        Caret caret = editor.getCaret();
        if (caret!=null)
            caret.addChangeListener(this);
        editor.addPropertyChangeListener(this);
        dragAndDropSupport.install(c);
    }
    public void deinstall(TextComponent c) {
        if (Debug.DEBUG && editor!=c)
            throw new InternalError();
        dragAndDropSupport.uninstall(c);
        editor.removePropertyChangeListener(this);
        Caret caret = editor.getCaret();
        if (caret!=null)
            caret.removeChangeListener(this);
        Document doc = editor.getDocument();
        if (doc!=null)
            doc.removeUndoableEditListener(this);
        ButtonManager bm = editor.getButtonManager();
        if (bm!=null)
            uninstallActions(bm);
        editor = null;
    }
    public abstract ViewFactory getViewFactory();
    public Action getDefaultAction() {
        return defaultAction;
    }
    public Caret createCaret() {
        return new DefaultCaret();
    }
    public abstract Document createDefaultDocument();

    /*DefaultEditorKit**********************************/
    public abstract Transferable copy(Document document,
                                      int startPos, int endPos)
        throws BadLocationException;
    public abstract boolean paste(Document document, int pos,
                                  Transferable transferable)
        throws UnsupportedFlavorException, IOException, BadLocationException;

    /*ChangeListener*********************************/
    public void stateChanged(ChangeEvent e) {
        currentFormat = null;
    }

    /***ClipboardOwner******************************/
    public void lostOwnership(Clipboard clipboard, Transferable contents) {}

    /*Actions***********************************************/
    protected final UndoAction undoActionInstance = new UndoAction();
    protected final RedoAction redoActionInstance = new RedoAction();
    private final Action[] actions = {
        new DeletePrevCharAction(), new DeleteNextCharAction(),
        new InsertContentAction(), 
        new PageUpAction(pageUpAction, false), 
        new PageDownAction(pageDownAction, false),
        new PageUpAction(selectionPageUpAction, true), 
        new PageDownAction(selectionPageDownAction, true),
        new InsertBreakAction(),
        new BeepAction(),
        new NextVisualPositionAction(forwardAction, false,
                                     SwingConstants.EAST),
        new NextVisualPositionAction(backwardAction, false,
                                     SwingConstants.WEST),
        new NextVisualPositionAction(selectionForwardAction, true,
                                     SwingConstants.EAST),
        new NextVisualPositionAction(selectionBackwardAction, true,
                                     SwingConstants.WEST),
        new NextVisualPositionAction(upAction, false,
                                     SwingConstants.NORTH),
        new NextVisualPositionAction(downAction, false,
                                     SwingConstants.SOUTH),
        new NextVisualPositionAction(selectionUpAction, true,
                                     SwingConstants.NORTH),
        new NextVisualPositionAction(selectionDownAction, true,
                                     SwingConstants.SOUTH),
        new BeginWordAction(beginWordAction, false),  
        new EndWordAction(endWordAction, false),
        new BeginWordAction(selectionBeginWordAction, true),  
        new EndWordAction(selectionEndWordAction, true),
        new PreviousWordAction(previousWordAction, false),  
        new NextWordAction(nextWordAction, false),
        new PreviousWordAction(selectionPreviousWordAction, true),  
        new NextWordAction(selectionNextWordAction, true),
        new BeginLineAction(beginLineAction, false),  
        new EndLineAction(endLineAction, false),
        new BeginLineAction(selectionBeginLineAction, true),  
        new EndLineAction(selectionEndLineAction, true),
        new BeginParagraphAction(beginParagraphAction, false),  
        new EndParagraphAction(endParagraphAction, false),
        new BeginParagraphAction(selectionBeginParagraphAction, true),  
        new EndParagraphAction(selectionEndParagraphAction, true),
        new BeginAction(beginAction, false), 
        new EndAction(endAction, false),
        new BeginAction(selectionBeginAction, true), 
        new EndAction(selectionEndAction, true),
        new DefaultKeyTypedAction(),
        new SelectWordAction(), 
        new SelectLineAction(),
        new SelectParagraphAction(), 
        new SelectAllAction(),
        new UnselectAction(),
        undoActionInstance, redoActionInstance,
        new ReplaceTextAction(),
        new EditCopyAction(),
        new EditCutAction(),
        new EditPasteAction(),
    };
    abstract class CursorAction extends AbstractAction {
        /*properties***********************************/
        private boolean select;
        /*Constructor**********************************/
        public CursorAction(String name, boolean _sel) { 
            super(name); select = _sel;
        }
        protected void setCaret(int pos) {
            final Caret caret = editor.getCaret();
            if (select)
                caret.moveDot(pos);
            else
                caret.setDot(pos);
        }
        protected void setDot(int pos) {
            final Caret caret = editor.getCaret();
            if (select)
                caret.moveDot(pos);
            else
                caret.setDot(pos);
        }
        /*DocumentAction********************************/
        public abstract void actionPerformed() throws BadLocationException;
        /*ActionListener*********************************/
        public final void actionPerformed(ActionEvent evt) {
            if (editor!=null) 
                try {
                    actionPerformed();
                } catch (BadLocationException e) {
                    beep();
                }
        }
    }
    protected abstract class DocumentAction extends AbstractAction {
        /*properties***************************/
        protected boolean modifying = true;
        /*Constructor**********************************/
        public DocumentAction(String name) { 
            super(name); 
        }
        public DocumentAction(String name, boolean _mod) { 
            super(name); modifying = _mod;
        }
        /*DocumentAction********************************/
        public abstract void actionPerformed(Document doc, int start, 
                                             int end, ActionEvent evt)
            throws BadLocationException;
        /*ActionListener*********************************/
        public final void actionPerformed(ActionEvent evt) {
            if (editor!=null) {
                if (modifying && (!editor.isEditable() || !editor.isEnabled())) {
                    beep(); return;
                }
                try {
                    final Caret caret = editor.getCaret();
                    final int dot = caret.getDot();
                    final int mark = caret.getMark();
                    actionPerformed(editor.getDocument(), 
                                    Math.min(dot, mark),
                                    Math.max(dot, mark), evt);
                } catch (BadLocationException e) {
                    beep(); return;
                }
            }
        }
    }
    protected abstract class InsertElementAction extends DocumentAction {
        public InsertElementAction(String name) {
            super(name);
        }
        protected abstract Element createElement(Document doc);
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt2) 
            throws BadLocationException {
            Element el = createElement(doc);
            if (el!=null) {
                Caret caret = editor.getCaret();
                try {
                    doc.beginUpdate();
                    DocumentEvent evt = new DocumentEvent(doc);
                    if (doc.getRootElement().pasteElementAt(caret.getDot(), el, evt))
                        doc.fireDocumentChanged(evt);
                } finally {
                    doc.endUpdate();
                }
            }
        }
    }
    protected abstract class EmbedElementAction extends DocumentAction {
        public EmbedElementAction(String name) {
            super(name);
        }
        protected abstract Element createElement(Document doc);
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt2) 
            throws BadLocationException {
            Element el = createElement(doc);
            if (el!=null) {
                try {
                    doc.beginUpdate();
                    DocumentEvent evt = new DocumentEvent(doc);
                    if (doc.getRootElement().embed(el, start, end, evt))
                        doc.fireDocumentChanged(evt);
                } finally {
                    doc.endUpdate();
                }
            }
        }
    }
    final class UndoAction extends DocumentAction {
        public UndoAction() {
            super(undoAction); setEnabled(false);
        }
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt) {
            doc.lockUndo();
            try {
                undo.undo();
            } catch (CannotUndoException e) {
                System.out.println("Unable to undo: " + e);
                e.printStackTrace();
            }
            doc.unlockUndo();
            update();
            redoActionInstance.update();
        }
        public void update() {
            setEnabled(undo.canUndo());
        }
    }
    final class RedoAction extends DocumentAction {
        public RedoAction() { 
            super(redoAction); setEnabled(false); 
        }
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt) {
            doc.lockUndo();
            try {
                undo.redo();
            } catch (CannotRedoException e) {
                System.out.println("Unable to redo: " + e);
                e.printStackTrace();
            }
            doc.unlockUndo();
            update();
            undoActionInstance.update();
        }
        public void update() {
            setEnabled(undo.canRedo());
        }
    }
    class DeletePrevCharAction extends DocumentAction {
        public DeletePrevCharAction() {
            super(deletePrevCharAction);
        }
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt) 
            throws BadLocationException {
            if (start==end) {
                if (start>0)
                    doc.remove(start-1, 1);
            } else {
                doc.remove(start, end-start);
            }
        }
    }
    class DeleteNextCharAction extends DocumentAction {
        public DeleteNextCharAction() {
            super(deleteNextCharAction); 
        }
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt) 
            throws BadLocationException {
            if (start==end)
                doc.remove(start, 1);
            else
                doc.remove(start, end-start);
        }
    }
    public class DefaultKeyTypedAction extends AbstractAction {
        public DefaultKeyTypedAction() { super(defaultKeyTypedAction); }
        public void actionPerformed(ActionEvent e) {
            String content = e.getActionCommand();
            final int mod = e.getModifiers();
            if (content!=null && content.length()>0 && 
                (mod & ActionEvent.ALT_MASK)==0) {
                final char c = content.charAt(0);
                if (!Character.isISOControl(c))
                    editor.replaceSelection(content, currentFormat);
            }
        }
    }
    public class InsertContentAction extends AbstractAction {
        public InsertContentAction() { super(insertContentAction); }
        public void actionPerformed(ActionEvent e) {
            String content = e.getActionCommand();
            if (content!=null)
                editor.replaceSelection(content, null);
            else
                beep();
        }
    }
    public class InsertBreakAction extends AbstractAction {
        public InsertBreakAction() {
            super(insertBreakAction);
        }
        public void actionPerformed(ActionEvent e) {
            editor.replaceSelection("\n", null);
        }
    }
    public class BeepAction extends AbstractAction {
        public BeepAction() { super(beepAction); }
        public void actionPerformed(ActionEvent e) {
            beep();
        }
    }
    class PageUpAction extends CursorAction {
        public PageUpAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed()
            throws BadLocationException {
            Rectangle visible = new Rectangle();
            editor.computeVisibleRect(visible);
            int scrollOffset = visible.y;
            visible.y -= visible.height;
            if (visible.y<0)
                visible.y = 0;
            scrollOffset = scrollOffset - visible.y;
            editor.scrollRectToVisible(visible);
            int selectedIndex = editor.getCaret().getDot();
            if (selectedIndex!=-1) {
                TextUI ui = editor.getUI();
                Rectangle r = ui.modelToView(editor, selectedIndex);
                r.y -= scrollOffset;
                selectedIndex = ui.viewToModel(editor, new Point(r.x,r.y));
                Document doc = (Document)editor.getDocument();
                final int len = doc.getRootElement().getLength();
                if (selectedIndex!=0 && selectedIndex>(len-1))
                    selectedIndex = len-1;
                if (selectedIndex<0)
                    selectedIndex = 0;
                setCaret(selectedIndex);
            }
        }
    }
    class PageDownAction extends CursorAction {
        PageDownAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            Rectangle visible = new Rectangle();
            editor.computeVisibleRect(visible);
            int scrollOffset = visible.y;
            visible.y += visible.height;
            if((visible.y+visible.height) > editor.getHeight())
                visible.y = (editor.getHeight() - visible.height);
            scrollOffset = visible.y - scrollOffset;
            editor.scrollRectToVisible(visible);
            int selectedIndex = editor.getCaret().getDot();
            if (selectedIndex!=-1) {
                TextUI ui = editor.getUI();
                Rectangle r = ui.modelToView(editor, selectedIndex);
                r.y += scrollOffset;
                selectedIndex = ui.viewToModel(editor, new Point(r.x,r.y));
                Document doc = (Document)editor.getDocument();
                final int len = doc.getRootElement().getLength();
                if (selectedIndex!=0 && selectedIndex>(len-1))
                    selectedIndex = len-1;
                if (selectedIndex<0)
                    selectedIndex = 0;
                setCaret(selectedIndex);
            }
        }
    }
    class NextVisualPositionAction extends CursorAction {
        private int direction;
        NextVisualPositionAction(String nm, boolean select, int direction) {
            super(nm, select);
            this.direction = direction;
        }
        public void actionPerformed() throws BadLocationException {
            Caret caret = editor.getCaret();
            int dot = caret.getDot();
            if (caret!=null &&
                (direction==SwingConstants.NORTH ||
                 direction==SwingConstants.SOUTH)) {
                Point p = caret.getMagicCaretPosition();
                if (p==null) {
                    Rectangle r = editor.getUI().modelToView(editor, dot);
                    p = new Point(r.x, r.y);
                    caret.setMagicCaretPosition(p);
                }
            }
            dot = editor.getUI().getNextVisualPositionFrom
                (editor, dot, direction);
            setDot(dot);
            if (direction==SwingConstants.EAST ||
                direction==SwingConstants.WEST)
                editor.getCaret().setMagicCaretPosition(null);
        }
    }
    class BeginWordAction extends CursorAction {
        BeginWordAction(String nm, boolean select) {
            super(nm, select);
        }
        public int getWordStart(TextComponent c, int offs) 
            throws BadLocationException {
            Document doc = c.getDocument();
            int lineStart = Math.max(0, offs-64);
            int lineEnd = Math.min(doc.getRootElement().getLength(), offs+64);
            Segment sg = new Segment();
            doc.getText(lineStart, lineEnd-lineStart, sg);
            if (sg.count>0) {
                String s = new String(sg.array, sg.offset, sg.count);
                BreakIterator words = BreakIterator.getWordInstance();
                words.setText(s);
                int wordPosition = offs - lineStart;
                if (wordPosition>=words.last())
                    wordPosition = words.last()-1;
                words.following(wordPosition);
                offs = lineStart + words.previous();
            }
            return offs;
        }
        public void actionPerformed() throws BadLocationException {
            int offs = editor.getCaret().getDot();
            int begOffs = getWordStart(editor, offs);
            setCaret(begOffs);
        }
    }
    class EndWordAction extends CursorAction {
        EndWordAction(String nm, boolean select) {
            super(nm, select);
        }
        public int getWordEnd(TextComponent c, int offs) 
            throws BadLocationException {
            Document doc = c.getDocument();
            int lineStart = Math.max(0, offs-64);
            int lineEnd = Math.min(doc.getRootElement().getLength(), offs+64);
            Segment sg = new Segment();
            doc.getText(lineStart, lineEnd-lineStart, sg);
            if (sg.count>0) {
                String s = new String(sg.array, sg.offset, sg.count);
                BreakIterator words = BreakIterator.getWordInstance();
                words.setText(s);
                int wordPosition = offs - lineStart;
                if (wordPosition>=words.last())
                    wordPosition = words.last()-1;
                offs = lineStart + words.following(wordPosition);
            }
            return offs;
        }
        public void actionPerformed() throws BadLocationException {
            int offs = editor.getCaret().getDot();
            int endOffs = getWordEnd(editor, offs);
            setCaret(endOffs);
        }
    }
    class PreviousWordAction extends CursorAction {
        PreviousWordAction(String nm, boolean select) {
            super(nm, select);
        }
        int getPrevWordInParagraph(Element line, int offs) throws BadLocationException {
            // check parameters
            if (line==null)
                throw new BadLocationException("No more words", offs);
            int lineStart = line.getStartOffset();
            int lineEnd = line.getEndOffset();
            if (offs<lineStart || offs>lineEnd)
                throw new BadLocationException("No more words", offs);
            // fetch text from document
            Segment sg = new Segment();
            line.getDocument().getText(lineStart, lineEnd-lineStart, sg);
            // initialize BreakIterator
            BreakIterator words = BreakIterator.getWordInstance();
            words.setText(new String(sg.array, sg.offset, sg.count));
            // check limits
            if ((offs-lineStart)>=sg.count)
                offs = lineStart+sg.count-1;
            if (words.following(offs-lineStart)==BreakIterator.DONE)
                words.last();
            // find boundary
            int wordPosition = words.previous();
            if (wordPosition==(offs-lineStart))
                wordPosition = words.previous();
            if (wordPosition==BreakIterator.DONE)
                return BreakIterator.DONE;
            char ch = sg.array[sg.offset+wordPosition];
            while (!Character.isLetterOrDigit(sg.array[sg.offset+wordPosition])) {
                wordPosition = words.previous();
                if (wordPosition==BreakIterator.DONE)
                    return BreakIterator.DONE;
            }
            return lineStart+wordPosition;
        }
        public int getPreviousWord(TextComponent c, int offs) 
            throws BadLocationException {
            int prevWord;
            Document doc = c.getDocument();
            Element line = doc.getParagraphElement(offs, Position.Bias.Backward);
            for (prevWord = getPrevWordInParagraph(line, offs);
                 prevWord == BreakIterator.DONE; 
                 prevWord = getPrevWordInParagraph(line, offs)) {
                offs = line.getStartOffset()-1;
                line = doc.getParagraphElement(offs, Position.Bias.Backward);
            }
            return prevWord;
        }
        public void actionPerformed() throws BadLocationException {
            int offs = editor.getCaret().getDot();
            offs = getPreviousWord(editor, offs);
            setCaret(offs);
        }
    }
    class NextWordAction extends CursorAction {
        NextWordAction(String nm, boolean select) {
            super(nm, select);
        }
        int getNextWordInParagraph(Element line, int offs, boolean first) 
            throws BadLocationException {
            // check parameters
            if (line==null)
                throw new BadLocationException("No more words", offs);
            final int lineStart = line.getStartOffset();
            final int lineEnd = line.getEndOffset();
            if (offs<lineStart || offs>=lineEnd)
                throw new BadLocationException("No more words", offs);
            // fetch text from document
            Segment sg = new Segment();
            line.getDocument().getText(lineStart, lineEnd-lineStart, sg);
            // initialize BreakIterator
            BreakIterator words = BreakIterator.getWordInstance();
            words.setText(new String(sg.array, sg.offset, sg.count));
            // check limits
            if ((offs-lineStart)>=sg.count)
                return BreakIterator.DONE;
            if (first && (words.first()==(offs-lineStart)) &&	
                !Character.isWhitespace(sg.array[sg.offset+words.first()])) {
                return offs;
            }
            // find boundary
            int wordPosition = words.following(offs-lineStart);
            if (wordPosition==BreakIterator.DONE || 
                wordPosition>=sg.count)
                return BreakIterator.DONE;
            while (!Character.isLetterOrDigit(sg.array[sg.offset+wordPosition])) {
                wordPosition = words.next();
                if (wordPosition==BreakIterator.DONE || wordPosition>=sg.count)
                    return BreakIterator.DONE;
            }
            return lineStart+wordPosition;
        }
        public int getNextWord(TextComponent c, int offs) 
            throws BadLocationException {
            int nextWord;
            Document doc = c.getDocument();
            Element line = doc.getParagraphElement(offs, Position.Bias.Forward);
            for (nextWord = getNextWordInParagraph(line, offs, false);
                 nextWord == BreakIterator.DONE; 
                 nextWord = getNextWordInParagraph(line, offs, true)) {
                offs = line.getEndOffset();
                line = doc.getParagraphElement(offs, Position.Bias.Forward);
            }
            return nextWord;
        }
        public void actionPerformed() throws BadLocationException {
            int offs = editor.getCaret().getDot();
            offs = getNextWord(editor, offs);
            setCaret(offs);
        }
    }
    class BeginLineAction extends CursorAction {
        BeginLineAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            TextUI ui = editor.getUI();
            final int offs = editor.getCaret().getDot();
            final int begOffs = ui.getRowStart(editor, offs);
            if (begOffs!=-1) {
                setCaret(begOffs);
                editor.getCaret().setMagicCaretPosition(null);
            }
        }
    }
    class EndLineAction extends CursorAction {
        EndLineAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            TextUI ui = editor.getUI();
            final int offs = editor.getCaret().getDot();
            final int endOffs = ui.getRowEnd(editor, offs);
            if (endOffs!=-1) {
                setCaret(endOffs);
                editor.getCaret().setMagicCaretPosition(null);
            }
        }
    }
    class BeginParagraphAction extends CursorAction {
        BeginParagraphAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            Document doc = editor.getDocument();
            int offs = editor.getCaret().getDot();
            Element elem = doc.getParagraphElement(offs, Position.Bias.Backward);
            offs = elem.getStartOffset();
            setCaret(offs);
        }
    }
    class EndParagraphAction extends CursorAction {
        EndParagraphAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            Document doc = editor.getDocument();
            int offs = editor.getCaret().getDot();
            Element elem = doc.getParagraphElement(offs, Position.Bias.Backward);
            offs = elem.getEndOffset();
            setCaret(offs);
        }
    }
    class BeginAction extends CursorAction {
        BeginAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            setCaret(0);
        }
    }

    class EndAction extends CursorAction {
        EndAction(String nm, boolean select) {
            super(nm, select);
        }
        public void actionPerformed() throws BadLocationException {
            Document doc = (Document)editor.getDocument();
            int dot = doc.getRootElement().getLength();
            setCaret(dot);
        }
    }
    class SelectWordAction extends AbstractAction {
        private Action start, end;
        SelectWordAction() {
            super(selectWordAction);
            start = new BeginWordAction("_", false);
            end = new EndWordAction("_", true);
        }
        public void actionPerformed(ActionEvent e) {
            start.actionPerformed(e);
            end.actionPerformed(e);
        }
    }
    class SelectLineAction extends AbstractAction {
        private Action start, end;
        SelectLineAction() {
            super(selectLineAction);
            start = new BeginLineAction("_", false);
            end = new EndLineAction("_", true);
        }
        public void actionPerformed(ActionEvent evt) {
            start.actionPerformed(evt);
            end.actionPerformed(evt);
        }
    }
    class SelectParagraphAction extends AbstractAction {
        private Action start, end;
        SelectParagraphAction() {
            super(selectParagraphAction);
            start = new BeginParagraphAction("_", false);
            end = new EndParagraphAction("_", true);
        }
        public void actionPerformed(ActionEvent evt) {
            start.actionPerformed(evt);
            end.actionPerformed(evt);
        }
    }
    class SelectAllAction extends AbstractAction {
        SelectAllAction() { super(selectAllAction); }
        public void actionPerformed(ActionEvent evt) {
            Document doc = editor.getDocument();
            final Caret caret = editor.getCaret();
            caret.setDot(0);
            caret.moveDot(doc.getRootElement().getLength());
        }
    }
    class UnselectAction extends AbstractAction {
        UnselectAction() { super(unselectAction); }
        public void actionPerformed(ActionEvent evt) {
            final Caret caret = editor.getCaret();
            caret.setDot(caret.getDot());
        }
    }
    class ReplaceTextAction extends AbstractAction {
        ReplaceDialog rd = new ReplaceDialog();
        ReplaceTextAction() { super(replaceTextAction); }
        public void actionPerformed(ActionEvent evt) {
            DialogInterfacePanel ip = new DialogInterfacePanel(null);
            rd.showDialog(ip, editor);
        }
    }
    /*Clipboard actions*****************************/
    final class EditCopyAction extends DocumentAction {
        public EditCopyAction() { super(copyAction, false); }
        public void actionPerformed(Document document,
                                    int startPos, int endPos,
                                    ActionEvent evt) 
            throws BadLocationException {
            if (startPos==endPos)
                return;

            Clipboard clipboard = getClipboard();
            if (clipboard==null)
                return;

            Transferable transferable = copy(document, startPos, endPos);
            if (transferable==null)
                return;

            clipboard.setContents(transferable, DefaultEditorKit.this);
        }
    }
    final class EditCutAction extends DocumentAction {
        protected EditCopyAction copy = new EditCopyAction();
        public EditCutAction() { super(cutAction); }
        public void actionPerformed(Document doc, int start, int end, ActionEvent evt) 
            throws BadLocationException {
            if (start==end) return;
            copy.actionPerformed(doc, start, end, evt);
            doc.remove(start, end-start);
        }
    }
    final class EditPasteAction extends DocumentAction {
        public EditPasteAction() { super(pasteAction); }
        public void actionPerformed(Document document, int start, int end,
                                    ActionEvent evt) 
            throws BadLocationException {
            Clipboard clipboard = getClipboard();
            if (clipboard==null)
                return;

            try {
                Transferable transferable = clipboard.getContents(this);
                if (transferable==null)
                    return;

                if (DUMP_CLIPBOARD) {
                    System.err.println("EditPasteAction: transferable = "+transferable);
                    DataFlavor[] dataFlavors = transferable.getTransferDataFlavors();
                    for (int z=0; z<dataFlavors.length; z++) {
                        System.err.println("dataFlavors["+z+"] = "+dataFlavors[z]);
                        System.err.println("    mimeType = "+dataFlavors[z].getMimeType());
                        System.err.println("    class = "+dataFlavors[z].getRepresentationClass());
                    }
                }

                try {
                    document.beginUpdate();
                    document.remove(start, end-start);
                    paste(document, start, transferable);
                } finally {
                    document.endUpdate();
                }
            } catch (UnsupportedFlavorException e) {
                Debug.report(e);
            } catch (IOException e) {
                Debug.report(e);
            }
        }
    }
}
