/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DefaultHighlighter.java,v 1.4 2001/08/20 09:23:39 max Exp $
 *
 */

package make.swing.text.view;

import java.util.Vector;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Color;
import javax.swing.text.BadLocationException;
import make.swing.text.TextComponent;
import make.swing.text.Document;
import make.swing.text.TextUI;

public class DefaultHighlighter implements Highlighter {
    /***properties********************************/
    private Vector highlights = new Vector();
    private TextComponent component;
    /***Highlighter*********************************/
    public void install(TextComponent c) {
	component = c;
	removeAllHighlights();
    }
    public void deinstall(TextComponent c) {
	component = null;
    }
    public Object addHighlight(int p0, int p1, Highlighter.HighlightPainter p) 
	throws BadLocationException {
	Document doc = (Document)component.getDocument();
	TextUI mapper = component.getUI();
	HighlightInfo i = new HighlightInfo(p0, p1, p);
	highlights.addElement(i);
	mapper.damageRange(component, p0, p1);
        return i;
    }
    public void removeHighlight(Object tag) {
	HighlightInfo lhi = (HighlightInfo)tag;
	if (lhi.width>0 && lhi.height>0)
	    component.repaint(lhi.x, lhi.y, lhi.width, lhi.height);
	highlights.removeElement(tag);
    }
    public void removeAllHighlights() {
	TextUI mapper = component.getUI();
	int len = highlights.size();
	if (len != 0) {
	    int minX = 0;
	    int minY = 0;
	    int maxX = 0;
	    int maxY = 0;
	    for (int i = 0; i < len; i++) {
		HighlightInfo info = (HighlightInfo)highlights.elementAt(i);
		minX = Math.min(minX, info.x);
		minY = Math.min(minY, info.y);
		maxX = Math.max(maxX, info.x + info.width);
		maxY = Math.max(maxY, info.y + info.height);
	    }
	    if (minX!=maxX && minY!=maxY)
		component.repaint(minX, minY, maxX - minX, maxY - minY);
	    highlights.removeAllElements();
	}
    }
    public void changeHighlight(Object tag, int p0, int p1) throws BadLocationException {
	Document doc = (Document)component.getDocument();
	TextUI mapper = component.getUI();
	HighlightInfo lhi = (HighlightInfo)tag;
	if (lhi.width>0 && lhi.height>0)
	    component.repaint(lhi.x, lhi.y, lhi.width, lhi.height);
	lhi.width = lhi.height = 0;
	lhi.p0 = p0;
	lhi.p1 = p1;
	mapper.damageRange(component, Math.min(p0, p1), Math.max(p0, p1));
    }
    public Highlighter.Highlight[] getHighlights() {
	Highlighter.Highlight[] h = new Highlighter.Highlight[highlights.size()];
	highlights.copyInto(h);
	return h;
    }
    public void paintHighlights(Graphics g, int p0, int p1,
				Shape viewBounds,
				TextComponent editor, View view) {
	for (int counter = highlights.size()-1; counter>=0; counter--) {
	    Object tag = highlights.elementAt(counter);
	    HighlightInfo lhi = (HighlightInfo)tag;
	    final int start = lhi.getStartOffset();
	    final int end = lhi.getEndOffset();
	    if ((p0<start && p1>start) || (p0>=start && p0<end))
		lhi.paintHighlights(g, p0, p1, viewBounds,
				    editor, view);
	}
    }
    /***Class**************************************/
    protected static final DefaultHighlightPainter defaultPainter
	= new DefaultHighlightPainter(null);
    public static Highlighter.HighlightPainter getDefaultPainter() {
	return defaultPainter;
    }
    public static class DefaultHighlightPainter implements Highlighter.HighlightPainter {
	/***properties********************/
	private Color color;
	/***Constructor**********************/
        public DefaultHighlightPainter(Color c) {
	    color = c;
	}
	/***Public*************************/
	public Color getColor() {
	    return color;
	}
	/***HighlightPainer************************/
	public Shape paint(Graphics g, int offs0, int offs1,
			   Shape bounds, TextComponent c, View view) {
	    Color color = getColor();
	    g.setColor(color==null?c.getSelectionColor():color);
	    if (offs0==view.getStartOffset() &&
		offs1==view.getEndOffset()) {
		Rectangle alloc = bounds instanceof Rectangle
		    ?(Rectangle)bounds
		    :bounds.getBounds();
		g.fillRect(alloc.x, alloc.y, alloc.width, alloc.height);
		return alloc;
	    } else {
		try {
                    Shape shape = view.modelToView(offs0, offs1, bounds);
                    Rectangle r = shape instanceof Rectangle
			?(Rectangle)shape
			:shape.getBounds();
                    g.fillRect(r.x, r.y, r.width, r.height);
                    return r;
		} catch (BadLocationException e) {
		    e.printStackTrace();
		}
	    }
	    return null;
	}
    }
    protected class HighlightInfo implements Highlighter.Highlight {
	/***properties*************************/
	public int p0, p1;
	public int x, y, width, height;
        public Highlighter.HighlightPainter painter;
	/***Constructor*************************/
	public HighlightInfo(int _p0, int _p1, Highlighter.HighlightPainter _p) {
	    p0 = _p0; p1 = _p1;
	    painter = _p;
	}
	/***Internal***************************/
	protected void union(Shape bounds) {
	    if (bounds==null)	return;
	    Rectangle alloc = bounds instanceof Rectangle
		?(Rectangle)bounds:bounds.getBounds();
	    if (width==0 || height==0) {
		x = alloc.x;
		y = alloc.y;
		width = alloc.width;
		height = alloc.height;
	    } else {
		width = Math.max(x + width, alloc.x + alloc.width);
		height = Math.max(y + height, alloc.y + alloc.height);
		x = Math.min(x, alloc.x);
		width -= x;
		y = Math.min(y, alloc.y);
		height -= y;
	    }
	}
	/***Highlight************************/
	public int getStartOffset() { return p0; }
	public int getEndOffset() { return p1; }
	public Highlighter.HighlightPainter getPainter() {
	    return painter;
	}
	public void paintHighlights(Graphics g, int p0, int p1,
				    Shape viewBounds, TextComponent editor,
				    View view) {
	    final int start = getStartOffset();
	    final int end = getEndOffset();
	    p0 = Math.max(start, p0);
	    p1 = Math.min(end, p1);
	    union(painter.paint(g, p0, p1, viewBounds, editor, view));
	}
    }
}
