/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: Debug.java,v 1.12 2001/10/30 16:50:47 max Exp $
 *
 */

package make.util;

import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

/**
 * Debug is an abstract class with some static properties
 * and methods which may help to speed up writing debugging code.
 */
public abstract class Debug {
    /*properties*****************************/
    /**
     * A debug flag that specifies whether to include
     * debugging code in the binary .class files.
     */
    public static final boolean DEBUG = true;
    /**
     * A debug flag that specifies whether to include
     * debugging code for memory leak debugging 
     * in the binary .class files.
     */
    public static final boolean LEAK_DEBUG = false;
    public static final DateFormat logDateFormat
        = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
    /*Debug*******************************/
    /**
     * Redirects STDERR to a file.
     */
    public static final void logErr(File file) throws IOException {
        FileOutputStream fos = new FileOutputStream(file);
        System.setErr(new PrintStream(fos));
        Date now = new Date();
        System.err.println(logDateFormat.format(now)+" - started log into file");
        System.err.println("\tHardware Architecture: '"+
                           System.getProperty("os.arch")+
                           "'");
        System.err.println("\tOperating System (OS): '"+
                           System.getProperty("os.name")+
                           "' version '"+
                           System.getProperty("os.version")+
                           "'");
        System.err.println("\tJava Virtual Machine (JVM): '"+
                           System.getProperty("java.vendor")+
                           "' version '"+
                           System.getProperty("java.version")+
                           "'");
    }
    /**
     * Tries to find the sub exception of the specified exception.
     * If there is no sub exception found, the method returns null.
     */
    public static final Throwable getSubException(Throwable t) {
        final String[] methodNames = {
            "getCause", "getException"
        };
        final Class cls = t.getClass();
        final Class[] paramClasses = {};
        final Object[] params = {};
        for (int z=0; z<methodNames.length; z++) {
            try {
                Method m = cls.getMethod(methodNames[z], paramClasses);
                return (Throwable)m.invoke(t, params);
            } catch (Exception e) {
            }
        }
        return null;
    }
    /**
     * Tries to find the innermost exception.
     */
    public static final Throwable getInnerException(Throwable t) {
        Throwable t2;
        do {
            t2 = t;
            t = getSubException(t2);
        } while (t!=null);
        return t2;
    }
    /**
     * Reports an exception to STDERR.
     * @param t the exception object
     * @param msg an optional message describing the circumstances
     */
    public static final void report(Throwable t, String msg) {
        Date now = new Date();
        if (msg==null) msg = t.toString();
        System.err.println(logDateFormat.format(now)+" - "+msg);
        t.printStackTrace(System.err);
        while (true) {
            t = getSubException(t);
            if (t==null) break;
            System.err.println("\tSUB "+t);
            t.printStackTrace(System.err);
        }
    }
    /**
     * Reports an exception to STDERR.
     * @param t the exception object
     */
    public static final void report(Throwable t) {
        report(t, null);
    }
}
