/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: Phonet.java,v 1.6 2001/10/30 16:48:24 max Exp $
 *
 */

package make.util;

/*
 *  This file is based on:
 *
 *  phonet.c from c't 25/99 (http://www.heise.de/ct)
 *
 *  Copyright (c):
 *  1999,2000:  Joerg MICHAEL, Adalbert-Stifter-Str. 11, 30655 Hannover, Germany
 *      and
 *  (version 1.0) 1999:  Heise Verlag, Helstorfer Str. 7, 30625 Hannover, Germany
 *
 *  The original souce code said:  
 *
 *  If you have any remarks, feel free to e-mail to:  
 *      ct@ct.heise.de
 *  The author's e-mail address is:  gerhard.michael@nienburg-weser.de
 *  (this address is correct, and yes, my name is  Joerg Michael).
 *
 *
 */

import java.util.Locale;
import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.BufferedReader;

/**
 * Phonet is a library that converts a specified string into
 * its phonetic code. It is similar to Donald Knuth's Soundex, 
 * but is better for languages like German.
 * <BR />
 * This algorithm was written by 
 * Joerg Michael (gerhard.michael@nienburg-weser.de)
 * and was published in c't 25/99 (http://www.heise.de/ct)
 * <BR />
 * Look at phonet_de.txt for an example of a German ruleset.
 */
public class Phonet {
    /*CONST*****************************/
    public static final int HASH_COUNT = 256;
    protected static final boolean TRACE = false;
    /*static*****************************/
    protected Hashtable phonets = new Hashtable();
    /*properties***************************/
    protected Locale locale;
    protected int[] hash1 = new int[HASH_COUNT];
    protected int[] hash2 = new int[HASH_COUNT];
    protected int[] hash;
    protected String[] expressions;
    protected String[] rules1;
    protected String[] rules2;
    protected String[] rules;
    /*Constructor*************************/
    protected Phonet(Locale _locale, Reader _rd) throws IOException {
        locale = _locale;
        BufferedReader rd = new BufferedReader(_rd);
        // clear hash
        for (int z=0; z<hash1.length; z++)
            hash1[z] = -1;
        for (int z=0; z<hash2.length; z++)
            hash2[z] = -1;
        // read rules from file
        Vector vc = new Vector();
        String line;
        for (int index=0; (line=rd.readLine())!=null;) {
            while (line.charAt(line.length()-1)=='\n' ||
                   line.charAt(line.length()-1)=='\r')
                line = line.substring(0, line.length()-1);
            // ignore comments
            if (line.length()==0 || line.charAt(0)=='#')
                continue;
            final int tab1 = line.indexOf('\t');
            final int tab2 = line.indexOf('\t', tab1+1);
            // extract fields
            String f1 = line.substring(0, tab1);
            String f2 = line.substring(tab1+1, tab2);
            if (f2.equals("null")) f2 = null;
            String f3 = line.substring(tab2+1);
            if (f3.equals("null")) f3 = null;
            vc.addElement(f1);
            vc.addElement(f2);
            vc.addElement(f3);
            // fill hash
            final int k = (int)f1.charAt(0);
            if (k>=0 && k<HASH_COUNT) {
                if (f2!=null && hash1[k]==-1)
                    hash1[k] = index;
                if (f3!=null && hash2[k]==-1)
                    hash2[k] = index;
            }
            index++;
        }
        final int ruleCount = vc.size()/3;
        expressions = new String[ruleCount];
        rules1 = new String[ruleCount];
        rules2 = new String[ruleCount];
        for (int z=0; z<ruleCount; z++) {
            expressions[z] = (String)vc.elementAt(z*3);
            rules1[z] = (String)vc.elementAt(z*3+1);
            rules2[z] = (String)vc.elementAt(z*3+2);
        }
        selectRuleset(0);
    }
    public void selectRuleset(int idx) {
        switch (idx) {
        case 0: rules = rules1; hash = hash1; break;
        case 1: rules = rules2; hash = hash2; break;
        default: throw new IllegalArgumentException("idx="+idx);
        }
    }
    protected int getHash(char c) {
        final int i = (int)c;
        return i>=0 && i<HASH_COUNT
            ?hash[i]
            :-1;
    }
    public String translate(String src) {
        src = src.toUpperCase(locale);
        char[] ca = src.toCharArray();
        final int len = 101;
        StringBuffer dest = new StringBuffer(len);
        int z = 0;
        for (int i=0; i<ca.length;) {
            char c = ca[i];
            if (TRACE)
                System.err.println("check position "+i+":  src = '"+src.substring(i)+"'");
            int n = getHash(c);
            if (n<0) {
                final int j = dest.length();
                if (j<len-1 && c!=0 && (j==0 || dest.charAt(j-1)!=c)) {
                    dest.append(c);
                }
                i++; z=0;
                continue;
            }
            boolean z0 = false;
            for (; n<expressions.length && expressions[n].charAt(0)==c; n++) {
                if (rules[n]==null)
                    continue;
                int matchCount = 1;
                int priority = 5;
                CursorString expression = new CursorString(expressions[n], 1);
                if (TRACE)
                    System.err.println("> rule no. "+n+"("+expression+") is being checked");
                while ((i+matchCount)<ca.length &&
                       ca[i+matchCount]==expression.current() &&
                       !Character.isDigit(expression.current()) && 
                       "(-<^$".indexOf(expression.current())==-1) {
                    matchCount++; expression.next();
                }
		/**  check an array of letters  ****/
                if (expression.current()=='(') {
                    if ((i+matchCount)<ca.length &&
                        Character.isLetter(ca[i+matchCount]) &&
                        expression.find(ca[i+matchCount])) {
                        matchCount++;
                        expression.seek(')');
                        expression.next();
                    }
                }
		// save current state
                final char p0 = expression.current();
                int matchCount0 = matchCount;
		// step back
                while (matchCount>1 && expression.current()=='-') {
                    // step back
                    matchCount--; expression.next();
                }
                if (expression.current()=='<')
                    expression.next();
                if (Character.isDigit(expression.current())) {
                    // read priority
                    priority = Integer.parseInt(""+expression.current());
                    expression.next();
                }
                if (expression.current()=='^' &&
                    expression.get(1)=='^')
                    expression.next();
                final char afterChar = (i+matchCount0)>=ca.length
                    ?0
                    :ca[i+matchCount0];
                if (expression.current()==0
                    ||
                    (expression.current()=='^' 
                     && 
                     (i==0 || !Character.isLetter(ca[i-1]))
                     &&
                     (expression.get(1)!='$' ||
                      (!Character.isLetter(afterChar) && afterChar!='.')))
                    ||
                    (expression.current()=='$' && 
                     i>0 && Character.isLetter(ca[i-1]) &&
                     !Character.isLetter(afterChar) && afterChar!='.')) {
                    /**  look for continuation, if:         ****/
                    /**  k > 1  and  NO '-' in first string ****/
                    final char c0 = ca[i+matchCount-1];
                    int n0 = getHash(c0);
                    if (matchCount>1 && n0>=0 &&
                        p0!='-' && (i+matchCount)<ca.length) {
                        int priority0 = 5;
                        for (; n0<expressions.length && 
                                 expressions[n0].charAt(0)==c0; n0++) {
                            if (rules[n0]==null)
                                continue;
                            if (TRACE)
                                System.err.println("> > continuation rule no. "+n0+" ("+new CursorString(expressions[n0], 1)+") is being checked");
                            /**  check whole string  ****/
                            matchCount0 = matchCount;
                            priority0 = 5;
                            expression = new CursorString(expressions[n0], 1);
                            while (expression.current()!=0 &&
                                   (i+matchCount0)<ca.length &&
                                   ca[i+matchCount0]==expression.current() &&
                                   !Character.isDigit(expression.current()) &&
                                   "(-<^$".indexOf(expression.current())==-1) {
                                matchCount0++; expression.next();
                            }
                            if (expression.current()=='(') {
                                /**  check an array of letters  ****/
                                if ((i+matchCount0)<ca.length &&
                                    Character.isLetter(ca[i+matchCount0]) &&
                                    expression.find(ca[i+matchCount0])) {
                                    matchCount0++;
                                    expression.seek(')');
                                    expression.next();
                                }
                            }
                            while (expression.current()=='-')
                                expression.next();
                            if (expression.current()=='<')
                                expression.next();
                            if (Character.isDigit(expression.current())) {
                                // read priority
                                priority0 = Integer.parseInt(""+expression.current());
                                expression.next();
                            }
                            /**  *s == '^' is not possible here  ****/
                            if (expression.current()==0 
                                ||
                                (expression.current()=='$' &&
                                 ((i+matchCount0)>=ca.length ||
                                  (!Character.isLetter(ca[i+matchCount0]) &&
                                   ca[i+matchCount0]!='.')))) {
                                if (matchCount0==matchCount) {
                                    if (TRACE)
                                        System.err.println("> > continuation rule no. "+n0+" not used (too short)");
                                    continue;
                                }
                                if (priority0<priority) {
                                    if (TRACE)
                                        System.err.println("> > continuation rule no. "+n0+" not used (too priority)");
                                    continue;
                                }
                                break;
                            }
                            if (TRACE)
                                System.err.println("> > continuation rule no. "+n0+" not used");
                        }
			
                        if (priority0>=priority && n0<expressions.length &&
                            expressions[n0].charAt(0)==c0) {
                            if (TRACE)
                                System.err.println("> > rule no. "+n+" not used because of continuation");
                            continue;
                        }
                    }
			
                    if (TRACE)
                        System.err.println("rule "+expressions[n]+" is applied");

                    CursorString rule = new CursorString(rules[n], 0);
                    int priority0 = expressions[n].charAt(0)!=0 &&
                        expressions[n].indexOf('<', 1)!=-1
                        ? 1 : 0;

                    if (priority0==1 && z==0) {
                        final int destLength = dest.length();
                        if (destLength>0 && rule.current()!=0 &&
                            (dest.charAt(destLength-1)==c || 
                             dest.charAt(destLength-1)==rule.current())) {
                            dest.setLength(destLength-1);
                        }
                        z0 = true;
                        z++;
                        int k0 = 0;
                        while (rule.current()!=0 && (i+k0)<ca.length) {
                            ca[i+k0] = rule.current();
                            k0++; rule.next();
                        }
                        if (matchCount>k0) {
                            src = new String(ca, 0, i+k0) 
                                + new String(ca, i+matchCount, ca.length-i-matchCount);
                            ca = src.toCharArray();
                        }
                        c = ca[i];
                    } else {
                        i += matchCount-1;
                        z = 0;
                        while (rule.current()!=0 && rule.get(1)!=0 && dest.length()<len-1) {
                            if (dest.length()==0 || dest.charAt(dest.length()-1)!=rule.current()) {
                                dest.append(rule.current());
                            }
                            rule.next();
                        }
                        c = rule.current();

                        if (expressions[n].length()>0 &&
                            expressions[n].indexOf("^^", 1)>0) {
                            if (c!=0) {
                                dest.append(c);
                            }
                            src = src.substring(i+1);
                            ca = src.toCharArray();
                            i = 0; z0 = true;
                        }
                    }
                    break;
                }
            }
            if (!z0) {
                final int j = dest.length();
                if (j<len-1 && c!=0 && (j==0 || dest.charAt(j-1)!=c)) {
                    dest.append(c);
                }
                i++; z=0;
            }
        }
        return dest.toString();
    }
    /*Static**************************/
    protected static InputStream openRuleStream(Locale locale) throws IOException {
        Class cls = Phonet.class;
        String fn;
        InputStream is;
        fn = "phonet_"+locale.getLanguage()+"_"+locale.getCountry()+"_"+locale.getVariant()+".txt";
        is = cls.getResourceAsStream(fn);
        if (is!=null) return is;
        fn = "phonet_"+locale.getLanguage()+"_"+locale.getCountry()+".txt";
        is = cls.getResourceAsStream(fn);
        if (is!=null) return is;
        fn = "phonet_"+locale.getLanguage()+".txt";
        is = cls.getResourceAsStream(fn);
        if (is!=null) return is;
        fn = "phonet.txt";
        is = cls.getResourceAsStream(fn);
        if (is!=null) return is;
        return null;
    }
    public static Phonet getPhonet(Locale locale) {
        try {
            InputStream is = openRuleStream(locale);
            if (is==null) return null;
            return new Phonet(locale, new InputStreamReader(is, "ISO-8859-1"));
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }
    }
    public static Phonet getPhonet() {
        return getPhonet(Locale.getDefault());
    }
}
