/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: ValueSortedMap.java,v 1.1 2001/12/19 16:57:21 max Exp $
 *
 */

package make.util;

import java.util.Collection;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.Comparator;

/**
 * This is a SortedMap implementation which sorts by the value,
 * rather than by key.
 */
public class ValueSortedMap
    implements SortedMap, Comparator {
    /*properties******************************/
    protected Map map = new HashMap();
    protected Comparator valueComparator;
    protected SortedMap sortedMap;

    /*Constructor**************************/
    public ValueSortedMap(Comparator _vc) {
        valueComparator = _vc;
        sortedMap = new TreeMap((Comparator)this);
    }

    /*Map**********************************/
    // Query Operations
    public int size() {
        return map.size();
    }
    public boolean isEmpty() {
        return map.isEmpty();
    }
    public boolean containsKey(Object key) {
        return map.containsKey(key);
    }
    public boolean containsValue(Object value) {
        return map.containsValue(value);
    }
    public Object get(Object key) {
        return map.get(key);
    }
    // Modification Operations
    public Object put(Object key, Object value) {
        map.put(key, value);
        return sortedMap.put(key, value);
    }
    public Object remove(Object key) {
        sortedMap.remove(key);
        return map.remove(key);
    }
    // Bulk Operations
    public void putAll(Map t) {
        Iterator it = t.entrySet().iterator();
        while (it.hasNext()) {
            Entry entry = (Entry)it.next();
            put(entry.getKey(), entry.getValue());
        }
    }
    public void clear() {
        sortedMap.clear();
        map.clear();
    }
    // Views
    public Set keySet() {
        return sortedMap.keySet();
    }
    public Collection values() {
        return sortedMap.values();
    }
    public Set entrySet() {
        return sortedMap.entrySet();
    }

    /*SortedMap****************************/
    public Comparator comparator() {
        return valueComparator;
    }
    public SortedMap subMap(Object fromKey, Object toKey) {
        return sortedMap.subMap(fromKey, toKey);
    }
    public SortedMap headMap(Object toKey) {
        return sortedMap.headMap(toKey);
    }
    public SortedMap tailMap(Object fromKey) {
        return sortedMap.tailMap(fromKey);
    }
    public Object firstKey() {
        return sortedMap.firstKey();
    }
    public Object lastKey() {
        return sortedMap.lastKey();
    }

    /*Comparator*************************/
    public int compare(Object o1, Object o2) {
        Object value1 = map.get(o1);
        Object value2 = map.get(o2);
        if (value1==null && value2==null)
            return 0;
        if (value1==null)
            return 1;
        if (value2==null)
            return -1;
        return valueComparator.compare(value1, value2);
    }

    /*Object******************************/
    public boolean equals(Object o) {
        if (o instanceof Comparator)
            return o==this;
        else
            return sortedMap.equals(o);
    }
    public int hashCode() {
        return sortedMap.hashCode();
    }
}
