/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: DOMWriter.java,v 1.9 2001/12/19 17:31:43 max Exp $
 *
 */

package make.xml;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.io.UnsupportedEncodingException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.w3c.dom.Element;
import make.util.Strings;

/**
 * This class can be used to serialize a DOM tree into a java.io.Writer.
 * Use it to save a DOM tree into a XML file.
 */
public class DOMWriter {
    /***static*******************************************/
    public static final int WIDTH = 79;
    public static final String DEFAULT_ENCODING = "UTF-8";
    /***properties***************************************/
    protected PrintWriter out;
    protected DOMWriterFormat format;
    protected boolean spaceAllowed = true;
    protected boolean cfuml = false;
    protected int left = WIDTH;
    protected StringBuffer indent = new StringBuffer();
    protected String encoding = DEFAULT_ENCODING;
    protected boolean preformatted = true;
    /***Constructor**************************************/
    public DOMWriter(PrintWriter wr, String _encoding) {
        out = wr;
        encoding = _encoding;
    }
    public DOMWriter(Writer wr, String _encoding) {
        this(new PrintWriter(wr), _encoding);
    }
    public DOMWriter(OutputStream os, String _encoding) 
        throws UnsupportedEncodingException {
        this(new OutputStreamWriter(os, _encoding), _encoding);
    }
    public void flush() {
        out.flush();
    }
    /***Props*******************************************/
    public void setFormat(DOMWriterFormat _f) {
        format = _f;
    }
    public void setWDDXWorkaround(boolean _cfuml) {
        cfuml = _cfuml;
    }
    /***Print*******************************************/
    protected void printNoBreak(String s) {
        out.print(s); 
        left -= s.length();
    }
    protected void print(String s) {
        if (s.length()>left && spaceAllowed)
            println();
        printNoBreak(s);
    }
    protected void println() {
        if (!cfuml) {
            out.println();
            out.print(indent.toString());
            left = WIDTH-indent.length();
        }
    }
    public void print(Node node) {
        if (node==null) return;
        int type = node.getNodeType();
        switch (type) {
        case Node.DOCUMENT_NODE: { // print document
            Document doc = (Document)node;
            String tag = doc.getDocumentElement().getTagName();
            indent.setLength(0);
            if (!cfuml) {
                out.print("<?xml version=\"1.0\"");
                if (encoding!=null)
                    out.print(" encoding=\""+encoding+"\"");
                out.println("?>");
                println();
		/*if (docType!=null) {
		  String systemId = docType.getSystemId();
		  if (systemId!=null) {
		  out.print("<!DOCTYPE "+tag);
		  String publicId = docType.getPublicId();
		  if (publicId==null)
		  out.print(" SYSTEM \"");
		  else
		  out.print(" PUBLIC \""+Strings.escapeXMLText(publicId)+"\" \"");
		  out.print(Strings.escapeXMLText(systemId));
		  out.print('\"');
		  out.println('>');
		  }
		  }*/
            }
            //println();
            NodeList nl = node.getChildNodes();
            for (int z=0; z<nl.getLength(); z++)
                print(nl.item(z));
            println();
            flush();
            spaceAllowed = true;
            break;
        }
	
        // print element with attributes
        case Node.ELEMENT_NODE: {
            final boolean flowing = preformatted ||
                format==null || format.isFlowing((Element)node);
            if (!flowing)
                println();
            printNoBreak("<");
            printNoBreak(node.getNodeName());
            NodeList nl = node.getChildNodes();
            Node[] children = new Node[nl.getLength()+1];
            for (int i=0,z=0; i<nl.getLength(); i++) {
                Node ch = nl.item(i);
                children[z++] = ch;
            }
            Attr attrs[] = sortAttributes(node.getAttributes());
            for (int i=0; i<attrs.length; i++) {
                Attr attr = attrs[i];
                String val = Strings.escapeXMLText(attr.getNodeValue());
                print(' '+attr.getNodeName()+"=\""+val+'"');
            }
            if (children[0]==null) {
                printNoBreak("/>");
            } else {
                boolean text1 = children[0].getNodeType()==Node.TEXT_NODE &&
                    children[1]==null &&
                    (left-4)>(children[0].getNodeValue().length()+
                              node.getNodeName().length());
                printNoBreak(">");
                if (!flowing) {
                    indent.append("  ");
                    if (!text1 &&
                        children[0].getNodeType()!=Node.ELEMENT_NODE)
                        println();
                }
                // print children
                final boolean oldPreformatted = preformatted;
                final boolean oldSpaceAllowed = spaceAllowed;
                preformatted = format==null || format.isPreformatted((Element)node);
                if (preformatted)
                    spaceAllowed = false;
                try {
                    for (int z=0; children[z]!=null; z++)
                        print(children[z]);
                } finally {
                    preformatted = oldPreformatted;
                    spaceAllowed = oldSpaceAllowed;
                }
                // end tag
                if (!flowing) {
                    indent.setLength(indent.length()-2);
                    if (!text1)
                        println();
                }
                print("</"+node.getNodeName()+">");
            }
            break;
        }
	
        // handle entity reference nodes
        case Node.ENTITY_REFERENCE_NODE: {
            print('&'+node.getNodeName()+';');
            break;
        }
	
        // print cdata sections
        case Node.CDATA_SECTION_NODE: {
            print("<![CDATA["+node.getNodeValue()+"]]>");
            break;
        }

        // print text
        case Node.TEXT_NODE: {
            StringBuffer word = new StringBuffer();
            String text = cfuml
                ?Strings.escapeWDDXText(node.getNodeValue(), preformatted)
                :Strings.escapeXMLText(node.getNodeValue(), preformatted);
            if (preformatted) {
                print(text);
            } else {
                boolean wasSpace = false;
                spaceAllowed = true;
                final char[] textc = text.toCharArray();
                for (int z=0; z<textc.length; z++) {
                    final char ch = textc[z];
                    if (Character.isWhitespace(ch)) {
                        if (!wasSpace) {
                            word.append(' ');
                            print(word.toString());
                            word.setLength(0);
                            wasSpace = true;
                        }
                    } else {
                        word.append(ch);
                        wasSpace = false;
                    }
                }
                if (word.length()>0)
                    print(word.toString());
                spaceAllowed = wasSpace;
            }
            break;
        }
	
        // print processing instruction
        case Node.PROCESSING_INSTRUCTION_NODE: {
            printNoBreak("<?"+node.getNodeName());
            String data = node.getNodeValue();
            if (data!=null && data.length()>0)
                printNoBreak(' '+data);
            printNoBreak("?>");
            break;
        }
        }
    }
    protected Attr[] sortAttributes(NamedNodeMap attrs) {
        int len = attrs==null ? 0 : attrs.getLength();
        Attr array[] = new Attr[len];
        for (int i=0; i<len; i++) {
            array[i] = (Attr)attrs.item(i);
        }
        for (int i=0; i<len-1; i++) {
            String name  = array[i].getNodeName();
            int index = i;
            for (int j=i+1; j<len; j++) {
                String curName = array[j].getNodeName();
                if (curName.compareTo(name)<0) {
                    name  = curName;
                    index = j;
                }
            }
            if (index!=i) {
                Attr temp = array[i];
                array[i] = array[index];
                array[index] = temp;
            }
        }
        return array;
    }
} 
