/*
 *
 *  MakeLib
 *  Copyright (c) 1997-2001 Max Kellermann (max.kellermann@epost.de)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation;
 *  version 2.1 of the License.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: JDBCUtil.java,v 1.5 2001/08/20 09:23:39 max Exp $
 *
 */

package make.xml;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.PreparedStatement;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Helper functions to convert ResultSet objects
 * into their DOM representation.
 */
public abstract class JDBCUtil {
    /*Columns************************************/
    public interface Column {
	Element extractValue(Document doc, ResultSet rs, String name, int col)
	    throws SQLException;
    }
    public static abstract class TextColumn implements Column {
	protected abstract String getValue(ResultSet rs, int col)
	    throws SQLException;
	public Element extractValue(Document doc, ResultSet rs, String name, int col)
	    throws SQLException {
	    String value = getValue(rs, col);
	    if (value==null) {
		return null;
	    } else {
		Element el = doc.createElement(name);
		el.appendChild(doc.createTextNode(value));
		return el;
	    }
	}
    }
    public static class StringColumn extends TextColumn {
	protected String getValue(ResultSet rs, int col)
	    throws SQLException {
	    String value = rs.getString(col);
	    return rs.wasNull() ? null : value;
	}
    }
    public static class BooleanColumn extends TextColumn {
	protected String getValue(ResultSet rs, int col)
	    throws SQLException {
	    byte value = rs.getByte(col);
	    if (rs.wasNull())
		return null;
	    return value==0
		?"FALSE"
		:"TRUE";
	}
    }
    public static final Column stringColumn = new StringColumn();
    public static final Column defaultColumn = stringColumn;
    /*ResultSet************************************/
    public static void appendRecord(Element dest, ResultSet rs, Column[] cols)
	throws SQLException {
	Document doc = dest.getOwnerDocument();
	ResultSetMetaData meta = rs.getMetaData();
	final int nColumns = meta.getColumnCount();
	for (int z=0; z<nColumns; z++) { 
	    Column c = cols==null || z>=cols.length ? null : cols[z];
	    if (c==null)
		c = defaultColumn;
	    Element el = c.extractValue(doc, rs, meta.getColumnName(z+1), z+1);
	    if (el!=null)
		dest.appendChild(el);
	}
    }
    public static Element recordToElement(Document doc, ResultSet rs, 
					  String tag, Column[] cols)
	throws SQLException {
	Element rec = doc.createElement(tag);
	appendRecord(rec, rs, cols);
	return rec;
    }
    public static Element resultSetToElement(Document doc, ResultSet rs, 
					     String tag1, String tag2,
					     Column[] cols) 
	throws SQLException {
	Element root = doc.createElement(tag1);
	ResultSetMetaData meta = rs.getMetaData();
	final int nColumns = meta.getColumnCount();
	final String[] columnNames = new String[nColumns];
	for (int z=0; z<nColumns; z++)
	    columnNames[z] = meta.getColumnName(z+1);
	while (rs.next())
	    root.appendChild(recordToElement(doc, rs, tag2, cols));
	return root;
    }
    public static Document resultSetToDocument(ResultSet rs, Column[] cols) 
	throws SQLException {
	Document doc = DOMUtil.createDocument();
	doc.appendChild(resultSetToElement(doc, rs, "resultset", "record", cols));
	return doc;
    }
    /*Query**************************************/
    public static Element recordToElement(Document doc, PreparedStatement ps, 
					  String tag, Column[] cols)
	throws SQLException {
	ResultSet rs = ps.executeQuery();
	return rs.next()
	    ?recordToElement(doc, rs, tag, cols)
	    :null;
    }
}
